# -*- coding: utf8 -*-

"""

   Copyright 2014-2022 Andreas Würl

   Licensed under the Apache License, Version 2.0 (the "License");
   you may not use this file except in compliance with the License.
   You may obtain a copy of the License at

       http://www.apache.org/licenses/LICENSE-2.0

   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   See the License for the specific language governing permissions and
   limitations under the License.

"""

import math
from abc import ABCMeta, abstractmethod

import pyproj
import shapely.geometry


class Geometry:
    """
    abstract base class for geometries
    """

    __metaclass__ = ABCMeta

    __slots__ = ('srid',)

    default_srid = 4326

    def __init__(self, srid=default_srid):
        self.srid = srid

    def get_srid(self):
        return self.srid

    def set_srid(self, srid):
        self.srid = srid

    @property
    @abstractmethod
    def env(self):
        pass


class Envelope(Geometry):
    """
    definition of a coordinate envelope
    """

    __slots__ = ('x_min', 'x_max', 'y_min', 'y_max')

    def __init__(self, x_min, x_max, y_min, y_max, srid=Geometry.default_srid):
        super().__init__(srid)
        self.x_min = x_min
        self.x_max = x_max
        self.y_min = y_min
        self.y_max = y_max

    @property
    def y_delta(self):
        return abs(self.y_max - self.y_min)

    @property
    def x_delta(self):
        return abs(self.x_max - self.x_min)

    def contains(self, point):
        return (self.x_min <= point.x <= self.x_max) and \
               (self.y_min <= point.y <= self.y_max)

    @property
    def env(self):
        return shapely.geometry.LinearRing(
            [(self.x_min, self.y_min), (self.x_min, self.y_max), (self.x_max, self.y_max),
             (self.x_max, self.y_min)])

    def __repr__(self):
        return 'Envelope(x: %.4f..%.4f, y: %.4f..%.4f)' % (
            self.x_min, self.x_max, self.y_min, self.y_max)


class Grid(Envelope):
    """ grid characteristics"""

    __slots__ = ['x_div', 'y_div', '__x_bin_count', '__y_bin_count']

    def __init__(self, x_min, x_max, y_min, y_max, x_div, y_div, srid=Geometry.default_srid):
        super().__init__(x_min, x_max, y_min, y_max, srid)
        self.x_div = x_div
        self.y_div = y_div
        self.__x_bin_count = None
        self.__y_bin_count = None

    def get_x_bin(self, x_pos):
        return int(math.ceil(float(x_pos - self.x_min) / self.x_div)) - 1

    def get_y_bin(self, y_pos):
        return int(math.ceil(float(y_pos - self.y_min) / self.y_div)) - 1

    @property
    def x_bin_count(self):
        if not self.__x_bin_count:
            self.__x_bin_count = self.get_x_bin(self.x_max) + 1
        return self.__x_bin_count

    @property
    def y_bin_count(self):
        if not self.__y_bin_count:
            self.__y_bin_count = self.get_y_bin(self.y_max) + 1
        return self.__y_bin_count

    def get_x_center(self, cell_index):
        return self.x_min + (cell_index + 0.5) * self.x_div

    def get_y_center(self, row_index):
        return self.y_min + (row_index + 0.5) * self.y_div

    def __repr__(self):
        return 'Grid(x: %.4f..%.4f (%.4f, #%d), y: %.4f..%.4f (%.4f, #%d))' % (
            self.x_min, self.x_max, self.x_div, self.x_bin_count,
            self.y_min, self.y_max, self.y_div, self.y_bin_count)


class GridFactory:
    WGS84 = pyproj.CRS(f"epsg:{Geometry.default_srid}")

    __slots__ = ['min_lon', 'max_lon', 'max_lat', 'min_lat', 'coord_sys', 'ref_lon', 'ref_lat', 'grid_data']

    def __init__(self, min_lon, max_lon, min_lat, max_lat, coord_sys, ref_lon=None, ref_lat=None):
        self.min_lon = min_lon
        self.max_lon = max_lon
        self.min_lat = min_lat
        self.max_lat = max_lat
        self.coord_sys = coord_sys
        self.ref_lon = ref_lon
        self.ref_lat = ref_lat

        self.grid_data = {}

    @staticmethod
    def fix_max(minimum, maximum, delta):
        return minimum + math.floor((maximum - minimum) / delta) * delta

    def get_for(self, base_length):
        if base_length not in self.grid_data:
            ref_lon = self.ref_lon if self.ref_lon else (self.min_lon + self.max_lon) / 2.0
            ref_lat = self.ref_lat if self.ref_lat else (self.min_lat + self.max_lat) / 2.0

            utm_x, utm_y = pyproj.Transformer.from_crs(self.WGS84, self.coord_sys) \
                .transform(ref_lat, ref_lon)
            lat_d, lon_d = pyproj.Transformer.from_crs(self.coord_sys, self.WGS84) \
                .transform(utm_x + base_length, utm_y + base_length)

            delta_lon = lon_d - ref_lon
            delta_lat = lat_d - ref_lat

            max_lon = self.fix_max(self.min_lon, self.max_lon, delta_lon)
            max_lat = self.fix_max(self.min_lat, self.max_lat, delta_lat)

            self.grid_data[base_length] = Grid(self.min_lon, max_lon, self.min_lat, max_lat,
                                               delta_lon, delta_lat,
                                               Geometry.default_srid)

        return self.grid_data[base_length]


class GridElement:
    """
    raster data entry
    """

    __slots__ = ['count', 'timestamp']

    def __init__(self, count, timestamp):
        self.count = count
        self.timestamp = timestamp

    def __gt__(self, other):
        return self.count > other.count

    def __repr__(self):
        return "GridElement(%d, %s)" % (self.count, str(self.timestamp))



# -*- coding: utf8 -*-

"""

   Copyright 2014-2016 Andreas Würl

   Licensed under the Apache License, Version 2.0 (the "License");
   you may not use this file except in compliance with the License.
   You may obtain a copy of the License at

       http://www.apache.org/licenses/LICENSE-2.0

   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   See the License for the specific language governing permissions and
   limitations under the License.

"""

import datetime
import time

from .. import db


def create_time_interval(minute_length, minute_offset):
    end_time = datetime.datetime.now(datetime.timezone.utc)
    end_time = end_time.replace(microsecond=0)
    end_time += datetime.timedelta(minutes=minute_offset)
    start_time = end_time - datetime.timedelta(minutes=minute_length)
    return db.query.TimeInterval(start_time, end_time)


class TimingState:
    __slots__ = ['statsd_client', 'reference_time', 'name', 'info_text']

    def __init__(self, name, statsd_client):
        self.statsd_client = statsd_client

        self.reference_time = time.time()
        self.name = name
        self.info_text = []

    def get_seconds(self, reference_time=None):
        return time.time() - (reference_time if reference_time else self.reference_time)

    def get_milliseconds(self, reference_time=None):
        return max(1, int(self.get_seconds(reference_time) * 1000))

    def reset_timer(self):
        self.reference_time = time.time()

    def log_timing(self, key, reference_time=None):
        self.statsd_client.timing(key, self.get_milliseconds(reference_time))

    def log_gauge(self, key, value):
        self.statsd_client.gauge(key, value)

    def log_incr(self, key):
        self.statsd_client.incr(key)

    def add_info_text(self, info_text):
        self.info_text += [info_text]



# -*- coding: utf8 -*-

"""

   Copyright 2014-2016 Andreas Würl

   Licensed under the Apache License, Version 2.0 (the "License");
   you may not use this file except in compliance with the License.
   You may obtain a copy of the License at

       http://www.apache.org/licenses/LICENSE-2.0

   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   See the License for the specific language governing permissions and
   limitations under the License.

"""

import datetime
import logging
import os
from abc import abstractmethod

try:
    from html.parser import HTMLParser
except ImportError:
    from HTMLParser import HTMLParser

from injector import inject

try:
    from requests import Session
except ImportError:
    class Session:
        pass

from .. import config, util


class TransportAbstract:
    @abstractmethod
    def read_lines(self, source_path, post_process=None):
        pass


class FileTransport(TransportAbstract):
    def read_lines(self, source_path, post_process=None):
        if os.path.isfile(source_path):
            with open(source_path) as data_file:
                for line in data_file:
                    yield line


class HttpFileTransport(FileTransport):
    TIMEOUT_SECONDS = 60

    logger = logging.getLogger(__name__)
    html_parser = HTMLParser()

    @inject
    def __init__(self, configuration: config.Config, session=None):
        self.config = configuration
        self.session = session if session else Session()

    def __del__(self):
        if self.session:
            try:
                self.logger.debug("close http session '%s'" % self.session)
                self.session.close()
            except ReferenceError:
                pass

    def read_lines(self, source_url, post_process=None):
        response = self.session.get(
            source_url,
            auth=(self.config.get_username(), self.config.get_password()),
            stream=True,
            timeout=self.TIMEOUT_SECONDS)

        if response.status_code != 200:
            self.logger.debug("http status %d for get '%s" % (response.status_code, source_url))
            return []

        return self.split_lines(post_process(response.content).splitlines() if post_process else response.iter_lines())

    def split_lines(self, lines):
        return (self.process_line(html_line) for html_line in lines)

    @staticmethod
    def process_line(line):
        return line.decode('utf8')


class BlitzortungDataPath:
    default_host_name = 'data'
    default_region = 1

    def __init__(self, base_path=None):
        self.data_path = os.path.join(
            (base_path if base_path else 'https://{host_name}.blitzortung.org'),
            'Data'
        )

    def build_path(self, sub_path, **kwargs):
        parameters = kwargs

        if 'host_name' not in parameters:
            parameters['host_name'] = self.default_host_name

        if 'region' not in parameters:
            parameters['region'] = self.default_region

        return os.path.join(self.data_path, sub_path).format(**parameters)


class BlitzortungDataPathGenerator:
    time_granularity = datetime.timedelta(minutes=10)
    url_path_format = '%Y/%m/%d/%H/%M.log'

    def get_paths(self, start_time, end_time=None):
        for interval_start_time in util.time_intervals(start_time, self.time_granularity, end_time):
            yield interval_start_time.strftime(self.url_path_format)



from .base import FileTransport, HttpFileTransport, BlitzortungDataPath, BlitzortungDataPathGenerator
from .raw_signal import RawSignalsBlitzortungDataProvider
from .station import StationsBlitzortungDataProvider
from .strike import StrikesBlitzortungDataProvider


def strikes():
    from .. import INJECTOR

    return INJECTOR.get(StrikesBlitzortungDataProvider)


def stations():
    from .. import INJECTOR

    return INJECTOR.get(StationsBlitzortungDataProvider)


def raw():
    from .. import INJECTOR

    return INJECTOR.get(RawSignalsBlitzortungDataProvider)



# -*- coding: utf8 -*-

"""

   Copyright 2014-2016 Andreas Würl

   Licensed under the Apache License, Version 2.0 (the "License");
   you may not use this file except in compliance with the License.
   You may obtain a copy of the License at

       http://www.apache.org/licenses/LICENSE-2.0

   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   See the License for the specific language governing permissions and
   limitations under the License.

"""

import time

from injector import inject
from twisted.internet.defer import gatherResults
from twisted.python import log

from .general import TimingState, create_time_interval
from .. import db, geom
from ..data import Timestamp


class StrikeState(TimingState):
    __slots__ = ['end_time']

    def __init__(self, statsd_client, end_time):
        super().__init__("strikes", statsd_client)
        self.end_time = end_time


class StrikeQuery:
    @inject
    def __init__(self, strike_query_builder: db.query_builder.Strike, strike_mapper: db.mapper.Strike):
        self.strike_query_builder = strike_query_builder
        self.strike_mapper = strike_mapper

    def create(self, id_or_offset, minute_length, minute_offset, connection, statsd_client):
        time_interval = create_time_interval(minute_length, minute_offset)
        state = StrikeState(statsd_client, Timestamp(time_interval.end))

        id_interval = db.query.IdInterval(id_or_offset) if id_or_offset > 0 else None
        order = db.query.Order('id')
        query = self.strike_query_builder.select_query(db.table.Strike.table_name, geom.Geometry.default_srid,
                                                       time_interval=time_interval, order=order,
                                                       id_interval=id_interval)

        strikes_result = connection.runQuery(str(query), query.get_parameters())
        strikes_result.addCallback(self.strike_build_results, state=state)
        return strikes_result, state

    def strike_build_results(self, query_result, state):
        state.add_info_text("query %.03fs #%d" % (state.get_seconds(), len(query_result)))
        state.log_timing('strikes.query')

        reference_time = time.time()
        end_time = state.end_time
        strikes = tuple(
            (
                (end_time - strike.timestamp).seconds,
                strike.x,
                strike.y,
                strike.altitude,
                strike.lateral_error,
                strike.amplitude,
                strike.station_count
            ) for strike in self.create_strikes(query_result))

        result = {'s': strikes}

        if strikes:
            result['next'] = query_result[-1][0] + 1

        state.add_info_text(", result %.03fs" % state.get_seconds(reference_time))
        state.log_timing('strikes.build_result', reference_time)
        return result

    def create_strikes(self, query_results):
        for result in query_results:
            yield self.strike_mapper.create_object(result)

    def combine_result(self, strikes_result, histogram_result, state):
        query = gatherResults([strikes_result, histogram_result], consumeErrors=True)
        query.addCallback(self.compile_strikes_result, state=state)
        query.addErrback(log.err)
        return query

    @staticmethod
    def compile_strikes_result(result, state):
        strikes_result = result[0]
        histogram_result = result[1]

        final_result = {
            't': state.end_time.strftime("%Y%m%dT%H:%M:%S"),
            'h': histogram_result
        }
        final_result.update(strikes_result)

        state.add_info_text(", total %.03fs" % state.get_seconds())
        state.log_timing('strikes.total')
        print(state.info_text)

        return final_result



# -*- coding: utf8 -*-

"""

   Copyright 2014-2016 Andreas Würl

   Licensed under the Apache License, Version 2.0 (the "License");
   you may not use this file except in compliance with the License.
   You may obtain a copy of the License at

       http://www.apache.org/licenses/LICENSE-2.0

   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   See the License for the specific language governing permissions and
   limitations under the License.

"""

import time

from injector import inject
from twisted.internet.defer import gatherResults
from twisted.python import log

from .general import create_time_interval, TimingState
from .. import db


class StrikeGridState(TimingState):
    __slots__ = ['grid_parameters', 'time_interval']

    def __init__(self, statsd_client, grid_parameters, time_interval):
        super().__init__("strikes_grid", statsd_client)
        self.grid_parameters = grid_parameters
        self.time_interval = time_interval


class StrikeGridQuery:
    @inject
    def __init__(self, strike_query_builder: db.query_builder.Strike):
        self.strike_query_builder = strike_query_builder

    def create(self, grid_parameters, minute_length, minute_offset, count_threshold, connection, statsd_client):
        time_interval = create_time_interval(minute_length, minute_offset)

        state = StrikeGridState(statsd_client, grid_parameters, time_interval)

        query = self.strike_query_builder.grid_query(db.table.Strike.table_name, grid_parameters,
                                                     time_interval=time_interval, count_threshold=count_threshold)

        grid_query = connection.runQuery(str(query), query.get_parameters())
        grid_query.addCallback(self.build_strikes_grid_result, state=state)
        grid_query.addErrback(log.err)
        return grid_query, state

    @staticmethod
    def build_strikes_grid_result(results, state):
        state.add_info_text("grid query %.03fs #%d %s" % (state.get_seconds(), len(results), state.grid_parameters))
        state.log_timing('strikes_grid.query')

        reference_time = time.time()
        x_bin_count = state.grid_parameters.x_bin_count
        y_bin_count = state.grid_parameters.y_bin_count
        end_time = state.time_interval.end
        strikes_grid_result = tuple(
            (
                result['rx'],
                y_bin_count - result['ry'],
                result['strike_count'],
                -(end_time - result['timestamp']).seconds
            ) for result in results if 0 <= result['rx'] < x_bin_count and 0 < result['ry'] <= y_bin_count
        )
        state.add_info_text(", result %.03fs" % state.get_seconds(reference_time))
        state.log_timing('strikes_grid.build_result', reference_time)

        return strikes_grid_result

    def combine_result(self, strike_grid_result, histogram_result, state):
        combined_result = gatherResults([strike_grid_result, histogram_result], consumeErrors=True)
        combined_result.addCallback(self.build_grid_response, state=state)
        combined_result.addErrback(log.err)

        return combined_result

    @staticmethod
    def build_grid_response(results, state):
        state.log_timing('strikes_grid.results')

        grid_data = results[0]
        histogram_data = results[1]

        state.log_gauge('strikes_grid.size', len(grid_data) if grid_data else 0)
        state.log_incr('strikes_grid')

        grid_parameters = state.grid_parameters
        end_time = state.time_interval.end
        duration = state.time_interval.duration
        response = {'r': grid_data, 'xd': round(grid_parameters.x_div, 6),
                    'yd': round(grid_parameters.y_div, 6),
                    'x0': round(grid_parameters.x_min, 4),
                    'y1': round(grid_parameters.y_max + grid_parameters.y_div, 4),
                    'xc': grid_parameters.x_bin_count,
                    'yc': grid_parameters.y_bin_count,
                    't': end_time.strftime("%Y%m%dT%H:%M:%S"),
                    'dt': duration.seconds,
                    'h': histogram_data}
        state.add_info_text(", total %.03fs" % state.get_seconds())
        state.log_timing('strikes_grid.total')
        print("".join(state.info_text))

        return response


class GlobalStrikeGridQuery:
    @inject
    def __init__(self, strike_query_builder: db.query_builder.Strike):
        self.strike_query_builder = strike_query_builder

    def create(self, grid_parameters, minute_length, minute_offset, count_threshold, connection, statsd_client):
        time_interval = create_time_interval(minute_length, minute_offset)

        state = StrikeGridState(statsd_client, grid_parameters, time_interval)

        query = self.strike_query_builder.global_grid_query(db.table.Strike.table_name, grid_parameters,
                                                            time_interval=time_interval,
                                                            count_threshold=count_threshold)

        grid_query = connection.runQuery(str(query), query.get_parameters())
        grid_query.addCallback(self.build_strikes_grid_result, state=state)
        grid_query.addErrback(log.err)
        return grid_query, state

    @staticmethod
    def build_strikes_grid_result(results, state):
        state.add_info_text(
            "global grid query %.03fs #%d %s" % (state.get_seconds(), len(results), state.grid_parameters))
        state.log_timing('global_strikes_grid.query')

        reference_time = time.time()
        end_time = state.time_interval.end
        global_strikes_grid_result = tuple(
            (
                result['rx'],
                -result['ry'] - 1,
                result['strike_count'],
                -(end_time - result['timestamp']).seconds
            ) for result in results
        )
        state.add_info_text(", result %.03fs" % state.get_seconds(reference_time))
        state.log_timing('globaL_strikes_grid.build_result', reference_time)

        return global_strikes_grid_result

    def combine_result(self, strike_grid_result, histogram_result, state):
        combined_result = gatherResults([strike_grid_result, histogram_result], consumeErrors=True)
        combined_result.addCallback(self.build_grid_response, state=state)
        combined_result.addErrback(log.err)

        return combined_result

    @staticmethod
    def build_grid_response(results, state):
        state.log_timing('global_strikes_grid.results')

        grid_data = results[0]
        histogram_data = results[1]

        state.log_gauge('global_strikes_grid.size', len(grid_data))
        state.log_incr('global_strikes_grid')

        grid_parameters = state.grid_parameters
        end_time = state.time_interval.end
        duration = state.time_interval.duration
        response = {'r': grid_data,
                    'xd': round(grid_parameters.x_div, 6),
                    'yd': round(grid_parameters.y_div, 6),
                    't': end_time.strftime("%Y%m%dT%H:%M:%S"),
                    'dt': duration.seconds,
                    'h': histogram_data}
        state.add_info_text(", total %.03fs" % state.get_seconds())
        state.log_timing('global_strikes_grid.total')
        print("".join(state.info_text))

        return response



# -*- coding: utf8 -*-

"""

   Copyright 2014-2016 Andreas Würl

   Licensed under the Apache License, Version 2.0 (the "License");
   you may not use this file except in compliance with the License.
   You may obtain a copy of the License at

       http://www.apache.org/licenses/LICENSE-2.0

   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   See the License for the specific language governing permissions and
   limitations under the License.

"""
import datetime
import logging

import math
from injector import inject

from . import mapper
from . import query
from . import query_builder
from .. import data
from .. import geom
from ..logger import get_logger_name

try:
    import psycopg2
    import psycopg2.pool
    import psycopg2.extras
    import psycopg2.extensions
except ImportError as e:
    from . import create_psycopg2_dummy

    psycopg2 = create_psycopg2_dummy()

from abc import ABCMeta, abstractmethod


class Base:
    """
    abstract base class for database access objects

    creation of database

    as user postgres:

    createuser -i -D -R -S -W -E -P blitzortung
    createdb -E utf8 -O blitzortung blitzortung
    createlang plpgsql blitzortung
    psql -f /usr/share/postgresql/10/contrib/postgis-2.4/postgis.sql -d blitzortung
    psql -f /usr/share/postgresql/12/contrib/postgis-3.0/postgis.sql -d blitzortung
    psql -f /usr/share/postgresql/10/contrib/postgis-2.4/spatial_ref_sys.sql -d blitzortung

    psql blitzortung

    GRANT SELECT ON spatial_ref_sys TO blitzortung;
    GRANT SELECT ON geometry_columns TO blitzortung;
    GRANT INSERT, DELETE ON geometry_columns TO blitzortung;
    CREATE EXTENSION "btree_gist";

    """
    __metaclass__ = ABCMeta

    default_timezone = datetime.timezone.utc

    def __init__(self, db_connection_pool):

        self.logger = logging.getLogger(get_logger_name(self.__class__))
        self.db_connection_pool = db_connection_pool

        self.schema_name = ""
        self.table_name = ""

        while True:
            self.conn = self.db_connection_pool.getconn()
            self.conn.cancel()
            try:
                self.conn.reset()
            except psycopg2.OperationalError:
                print("reconnect to db")
                self.db_connection_pool.putconn(self.conn, close=True)
                continue
            break
        psycopg2.extensions.register_type(psycopg2.extensions.UNICODE, self.conn)
        psycopg2.extensions.register_type(psycopg2.extensions.UNICODEARRAY, self.conn)
        self.conn.set_client_encoding('UTF8')

        self.srid = geom.Geometry.default_srid
        self.tz = None
        self.set_timezone(Base.default_timezone)

        cur = None
        try:
            cur = self.conn.cursor(cursor_factory=psycopg2.extras.DictCursor)
        except psycopg2.DatabaseError as error:
            self.logger.error(error)

            if self.conn:
                try:
                    self.conn.close()
                except NameError:
                    pass
        finally:
            if cur:
                cur.close()

    def __del__(self):
        try:
            if not self.conn.closed:
                self.db_connection_pool.putconn(self.conn)
        except (psycopg2.pool.PoolError, AttributeError):
            pass

    def is_connected(self):
        if self.conn:
            return not self.conn.closed
        else:
            return False

    @property
    def full_table_name(self):
        if self.schema_name:
            return '"' + self.schema_name + '"."' + self.table_name + '"'
        else:
            return self.table_name

    def get_srid(self):
        return self.srid

    def set_srid(self, srid):
        self.srid = srid

    def get_timezone(self):
        return self.tz

    def set_timezone(self, tz):
        self.tz = tz
        with self.conn.cursor() as cur:
            cur.execute('SET TIME ZONE \'%s\'' % str(self.tz))

    def fix_timezone(self, timestamp):
        return timestamp.astimezone(self.tz) if timestamp else None

    def from_bare_utc_to_timezone(self, utc_time):
        return utc_time.replace(tzinfo=datetime.timezone.utc).astimezone(self.tz)

    @staticmethod
    def from_timezone_to_bare_utc(time_with_tz):
        return time_with_tz.astimezone(datetime.timezone.utc).replace(tzinfo=None)

    def commit(self):
        """ commit pending database transaction """
        self.conn.commit()

    def rollback(self):
        """ rollback pending database transaction """
        self.conn.rollback()

    @abstractmethod
    def insert(self, *args):
        pass

    @abstractmethod
    def select(self, **kwargs):
        pass

    def execute(self, sql_statement, parameters=None, factory_method=None, **factory_method_args):
        with self.conn.cursor(cursor_factory=psycopg2.extras.DictCursor) as cursor:
            cursor.execute(sql_statement, parameters)
            if factory_method:
                method = factory_method(cursor, **factory_method_args)
                return method

    def execute_single(self, sql_statement, parameters=None, factory_method=None, **factory_method_args):
        def single_cursor_factory(cursor):
            if cursor.rowcount == 1:
                return factory_method(cursor.fetchone(), **factory_method_args)

        return self.execute(sql_statement, parameters, single_cursor_factory)

    def execute_many(self, sql_statement, parameters=None, factory_method=None, **factory_method_args):
        with self.conn.cursor(cursor_factory=psycopg2.extras.DictCursor) as cursor:
            cursor.execute(sql_statement, parameters)
            if factory_method:
                for value in cursor:
                    yield factory_method(value, **factory_method_args)


class Strike(Base):
    """
    strike db access class

    database table creation (as db user blitzortung, database blitzortung):

    > psql -h localhost -U blitzortung -W blitzortung

    CREATE TABLE strikes (id bigserial, "timestamp" timestamptz, nanoseconds SMALLINT, geog GEOGRAPHY(Point),
        PRIMARY KEY(id));
    ALTER TABLE strikes ADD COLUMN altitude SMALLINT;
    ALTER TABLE strikes ADD COLUMN region SMALLINT;
    ALTER TABLE strikes ADD COLUMN amplitude REAL;
    ALTER TABLE strikes ADD COLUMN error2d SMALLINT;
    ALTER TABLE strikes ADD COLUMN stationcount SMALLINT;

    CREATE INDEX strikes_timestamp ON strikes USING btree("timestamp");
    CREATE INDEX strikes_region_timestamp_nanoseconds ON strikes USING btree(region, "timestamp", nanoseconds);
    CREATE INDEX strikes_id_timestamp ON strikes USING btree(id, "timestamp");
    CREATE INDEX strikes_geog ON strikes USING gist(geog);
    CREATE INDEX strikes_timestamp_geog ON strikes USING gist("timestamp", geog);
    CREATE INDEX strikes_id_timestamp_geog ON strikes USING gist(id, "timestamp", geog);

    empty the table with the following commands:

    DELETE FROM strikes;
    ALTER SEQUENCE strikes_id_seq RESTART 1;

    """

    table_name = 'strikes'

    @inject
    def __init__(self, db_connection_pool: psycopg2.pool.ThreadedConnectionPool, query_builder_: query_builder.Strike,
                 strike_mapper: mapper.Strike):
        super().__init__(db_connection_pool)

        self.table_name = Strike.table_name
        self.query_builder = query_builder_
        self.strike_mapper = strike_mapper

    def insert(self, strike, region=1):
        sql = 'INSERT INTO ' + self.full_table_name + \
              ' ("timestamp", nanoseconds, geog, altitude, region, amplitude, error2d, stationcount) ' + \
              'VALUES (%(timestamp)s, %(nanoseconds)s, ST_MakePoint(%(longitude)s, %(latitude)s), ' + \
              '%(altitude)s, %(region)s, %(amplitude)s, %(error2d)s, %(stationcount)s)'

        parameters = {
            'timestamp': strike.timestamp.datetime,
            'nanoseconds': strike.timestamp.nanosecond,
            'longitude': strike.x,
            'latitude': strike.y,
            'altitude': strike.altitude,
            'region': region,
            'amplitude': strike.amplitude,
            'error2d': strike.lateral_error,
            'stationcount': strike.station_count
        }

        self.execute(sql, parameters)

    def get_latest_time(self, region=None):
        sql = 'SELECT "timestamp", nanoseconds FROM ' + self.full_table_name + \
              (' WHERE region=%(region)s' if region else '') + \
              ' ORDER BY "timestamp" DESC, nanoseconds DESC LIMIT 1'

        def prepare_result(result):
            return data.Timestamp(self.fix_timezone(result['timestamp']), result['nanoseconds'])

        parameters = {'region': region}
        return self.execute_single(sql, parameters, prepare_result)

    def select(self, **kwargs):
        """ build up query """

        query_ = self.query_builder.select_query(self.full_table_name, self.srid, **kwargs)

        return self.execute_many(str(query_), query_.get_parameters(), self.strike_mapper.create_object,
                                 timezone=self.tz)

    def select_grid(self, grid, count_threshold, **kwargs):
        """ build up raster query """

        query = self.query_builder.grid_query(self.table_name, grid, count_threshold, **kwargs)

        def prepare_results(cursor, _):
            raster_data = data.GridData(grid)

            for result in cursor:
                raster_data.set(result['rx'], result['ry'],
                                geom.GridElement(result['count'], result['timestamp']))

            return raster_data

        return self.execute_base(str(query), query.get_parameters(), prepare_results)

    def select_histogram(self, minutes, minute_offset=0, binsize=5, region=None, envelope=None):

        query = self.query_builder.histogram_query(
            self.full_table_name,
            minutes, minute_offset, binsize,
            region, envelope
        )

        def prepare_result(cursor, _):
            value_count = minutes / binsize

            result = [0] * value_count

            for bin_data in cursor:
                result[bin_data[0] + value_count - 1] = bin_data[1]

            return result

        return self.execute(str(query), query.get_parameters(), prepare_result)


class Station(Base):
    """

    database table creation (as db user blitzortung, database blitzortung):

    CREATE TABLE stations (id bigserial, number int, "user" int, geog GEOGRAPHY(Point), PRIMARY KEY(id));
    ALTER TABLE stations ADD COLUMN region SMALLINT;
    ALTER TABLE stations ADD COLUMN name CHARACTER VARYING;
    ALTER TABLE stations ADD COLUMN country CHARACTER VARYING;
    ALTER TABLE stations ADD COLUMN "timestamp" TIMESTAMPTZ;

    CREATE INDEX stations_timestamp ON stations USING btree("timestamp");
    CREATE INDEX stations_number_timestamp ON stations USING btree(number, "timestamp");
    CREATE INDEX stations_geog ON stations USING gist(geog);

    empty the table with the following commands:

    DELETE FROM stations;
    ALTER SEQUENCE stations_id_seq RESTART 1;
    """

    @inject
    def __init__(self, db_connection_pool: psycopg2.pool.ThreadedConnectionPool, station_mapper: mapper.Station):
        super().__init__(db_connection_pool)

        self.table_name = 'stations'
        self.station_mapper = station_mapper

    def insert(self, station, region=1):
        self.execute('INSERT INTO ' + self.full_table_name +
                     ' (number, "user", "name", country, "timestamp", geog, region) ' +
                     'VALUES (%s, %s, %s, %s, %s, ST_MakePoint(%s, %s), %s)',
                     (station.number, station.user, station.name,
                      station.country, station.timestamp.datetime, station.x, station.y, region))

    def select(self, timestamp=None, region=None):
        sql = ''' select
             o.begin, s.number, s.user, s.name, s.country, s.geog
        from stations as s
        inner join
           (select b. region, b.number, max(b."timestamp") as "timestamp"
            from stations as b
        group by region, number
        order by region, number) as c
        on s.region = c.region and s.number = c.number and s."timestamp" = c."timestamp"
        left join stations_offline as o
        on o.number = s.number and o.region = s.region and o."end" is null'''

        if region:
            sql += ''' where s.region = %(region)s'''

        sql += ''' order by s.number'''

        return self.execute_many(sql, {'region': region}, self.station_mapper.create_object)


class StationOffline(Base):
    """

    database table creation (as db user blitzortung, database blitzortung):

    CREATE TABLE stations_offline (id bigserial, number int, PRIMARY KEY(id));
    ALTER TABLE stations_offline ADD COLUMN region SMALLINT;
    ALTER TABLE stations_offline ADD COLUMN begin TIMESTAMPTZ;
    ALTER TABLE stations_offline ADD COLUMN "end" TIMESTAMPTZ;

    CREATE INDEX stations_offline_begin ON stations_offline USING btree(begin);
    CREATE INDEX stations_offline_end ON stations_offline USING btree("end");
    CREATE INDEX stations_offline_end_number ON stations_offline USING btree("end", number);
    CREATE INDEX stations_offline_begin_end ON stations_offline USING btree(begin, "end");

    empty the table with the following commands:

    DELETE FROM stations_offline;
    ALTER SEQUENCE stations_offline_id_seq RESTART 1;
    """

    @inject
    def __init__(self, db_connection_pool: psycopg2.pool.ThreadedConnectionPool,
                 station_offline_mapper: mapper.StationOffline):
        super().__init__(db_connection_pool)

        self.table_name = 'stations_offline'
        self.station_offline_mapper = station_offline_mapper

    def insert(self, station_offline, region=1):
        self.execute('INSERT INTO ' + self.full_table_name +
                     ' (number, region, begin, "end") ' +
                     'VALUES (%s, %s, %s, %s)',
                     (station_offline.number, region, station_offline.begin.datetime,
                      station_offline.end.datetime if station_offline.end else None))

    def update(self, station_offline, region=1):
        self.execute('UPDATE ' + self.full_table_name + ' SET "end"=%s WHERE id=%s and region=%s',
                     (station_offline.end, station_offline.id, region))

    def select(self, timestamp=None, region=1):
        sql = '''select id, number, region, begin, "end"
            from stations_offline where "end" is null and region=%s order by number;'''

        return self.execute_many(sql, (region,), self.station_offline_mapper.create_object)


class Location(Base):
    """
    geonames db access class

    CREATE SCHEMA geo;

    CREATE TABLE geo.geonames (id bigserial, "name" character varying, geog Geography(Point), PRIMARY KEY(id));

    ALTER TABLE geo.geonames ADD COLUMN "class" INTEGER;
    ALTER TABLE geo.geonames ADD COLUMN feature_class CHARACTER(1);
    ALTER TABLE geo.geonames ADD COLUMN feature_code VARCHAR;
    ALTER TABLE geo.geonames ADD COLUMN country_code VARCHAR;
    ALTER TABLE geo.geonames ADD COLUMN admin_code_1 VARCHAR;
    ALTER TABLE geo.geonames ADD COLUMN admin_code_2 VARCHAR;
    ALTER TABLE geo.geonames ADD COLUMN population INTEGER;
    ALTER TABLE geo.geonames ADD COLUMN elevation SMALLINT;

    CREATE INDEX geonames_geog ON geo.geonames USING gist(geog);

    """

    @inject
    def __init__(self, db_connection_pool: psycopg2.pool.ThreadedConnectionPool):
        super().__init__(db_connection_pool)
        self.schema_name = 'geo'
        self.table_name = 'geonames'
        self.center = None
        self.min_population = None
        self.limit = None
        self.max_distance = None

    def delete_all(self):
        self.execute('DELETE FROM ' + self.full_table_name)

    def insert(self, line):
        fields = line.strip().split('\t')
        name = fields[1]
        y = float(fields[4])
        x = float(fields[5])
        feature_class = fields[6]
        feature_code = fields[7]
        country_code = fields[8]
        admin_code_1 = fields[10]
        admin_code_2 = fields[11]
        population = int(fields[14])
        if fields[15] != '':
            elevation = int(fields[15])
        else:
            elevation = -1

        name = name.replace("'", "''")

        classification = self.determine_size_class(population)

        if classification is not None:
            self.execute('INSERT INTO ' + self.full_table_name +
                         '(geog, name, class, feature_class, feature_code, country_code, admin_code_1, admin_code_2, ' +
                         'population, elevation)' +
                         'VALUES(ST_GeomFromText(\'POINT(%s %s)\', 4326), %s, %s, %s, %s, %s, %s, %s, %s, %s)',
                         (x, y, name, classification, feature_class, feature_code, country_code, admin_code_1,
                          admin_code_2, population, elevation))

    @staticmethod
    def determine_size_class(n):
        if n < 1:
            return None
        base = math.floor(math.log(n) / math.log(10)) - 1
        relative = n / math.pow(10, base)
        order = min(2, math.floor(relative / 25))
        if base < 0:
            base = 0
        return min(15, base * 3 + order)

    def create_object_instance(self, result):
        pass

    def select(self, *args):
        self.center = None
        self.min_population = 1000
        self.max_distance = 10000
        self.limit = 10

        for arg in args:
            if arg:
                if isinstance(arg, query.Center):
                    self.center = arg
                elif isinstance(arg, query.Limit):
                    self.limit = arg

        if self.is_connected():
            query_string = '''SELECT
                name,
                country_code,
                admin_code_1,
                admin_code_2,
                feature_class,
                feature_code,
                elevation,
                ST_Transform(geog::geometry, %(srid)s) AS geog,
                population,
                ST_Distance_Sphere(geog::geometry, c.center) AS distance,
                ST_Azimuth(geog::geometry, c.center) AS azimuth
            FROM
                (SELECT ST_SetSRID(ST_MakePoint(%(center_x)s, %(center_y)s), %(srid)s) as center ) as c,''' + \
                           self.full_table_name + '''
            WHERE
                feature_class='P'
                AND population >= %(min_population)s
                AND ST_Transform(geog::geometry, %(srid)s) && ST_Expand(c.center, %(max_distance)s)
            ORDER BY distance
            LIMIT %(limit)s'''

            params = {
                'srid': self.srid,
                'center_x': self.center.get_point().x,
                'center_y': self.center.get_point().y,
                'min_population': self.min_population,
                'max_distance': self.max_distance,
                'limit': self.limit
            }

            def build_results(cursor, _):
                locations = tuple(
                    {
                        'name': result['name'],
                        'distance': result['distance'],
                        'azimuth': result['azimuth']
                    } for result in cursor
                )

                return locations

            return self.execute_many(query_string, params, build_results)


class ServiceLogBase(Base):
    """

    Base class for servicelog tables
    """

    def get_latest_time(self):
        sql = 'SELECT "timestamp" FROM ' + self.full_table_name + \
              ' ORDER BY "timestamp" DESC LIMIT 1'

        def prepare_result(cursor, _):
            if cursor.rowcount == 1:
                result = cursor.fetchone()
                return data.Timestamp(self.fix_timezone(result['timestamp']))
            else:
                return None

        return self.execute(sql, factory_method=prepare_result)

    def select(self, args):
        pass

    def create_object_instance(self, result):
        pass


class ServiceLogTotal(ServiceLogBase):
    """
        CREATE TABLE servicelog_total ("timestamp" TIMESTAMPTZ, count INT);

        CREATE INDEX servicelog_total_timestamp ON servicelog_total USING btree("timestamp");
    """

    @inject
    def __init__(self, db_connection_pool: psycopg2.pool.ThreadedConnectionPool):
        super().__init__(db_connection_pool)

        self.table_name = 'servicelog_total'

    def insert(self, timestamp, count):
        sql = 'INSERT INTO ' + self.full_table_name + ' ' + \
              '("timestamp", count)' + \
              'VALUES (%(timestamp)s, %(count)s);'

        parameters = {
            'timestamp': timestamp,
            'count': count
        }

        self.execute(sql, parameters)


class ServiceLogCountry(ServiceLogBase):
    """
        CREATE TABLE servicelog_country ("timestamp" TIMESTAMPTZ, country_code CHARACTER VARYING, "count" INT);

        CREATE INDEX servicelog_country_timestamp ON servicelog_country USING btree("timestamp");
    """

    @inject
    def __init__(self, db_connection_pool: psycopg2.pool.ThreadedConnectionPool):
        super().__init__(db_connection_pool)

        self.table_name = 'servicelog_country'

    def insert(self, timestamp, country_code, count):
        sql = 'INSERT INTO ' + self.full_table_name + ' ' + \
              '("timestamp", country_code, "count")' + \
              'VALUES (%(timestamp)s, %(country_code)s, %(count)s);'

        parameters = {
            'timestamp': timestamp,
            'country_code': country_code,
            'count': count
        }

        self.execute(sql, parameters)


class ServiceLogVersion(ServiceLogBase):
    """
        CREATE TABLE servicelog_version ("timestamp" TIMESTAMPTZ, version CHARACTER VARYING, "count" INT);

        CREATE INDEX servicelog_version_timestamp ON servicelog_version USING btree("timestamp");
    """

    @inject
    def __init__(self, db_connection_pool: psycopg2.pool.ThreadedConnectionPool):
        super().__init__(db_connection_pool)

        self.table_name = 'servicelog_version'

    def insert(self, timestamp, version, count):
        sql = 'INSERT INTO ' + self.full_table_name + ' ' + \
              '("timestamp", version, "count")' + \
              'VALUES (%(timestamp)s, %(version)s, %(count)s);'

        parameters = {
            'timestamp': timestamp,
            'version': version,
            'count': count
        }

        self.execute(sql, parameters)


class ServiceLogParameters(ServiceLogBase):
    """
        CREATE TABLE servicelog_parameters ("timestamp" TIMESTAMPTZ, region INT, minute_length INT, minute_offset INT, grid_baselength INT, count_threshold INT, "count" INT);

        CREATE INDEX servicelog_parameters_timestamp ON servicelog_parameters USING btree("timestamp");
    """

    @inject
    def __init__(self, db_connection_pool: psycopg2.pool.ThreadedConnectionPool):
        super().__init__(db_connection_pool)

        self.table_name = 'servicelog_parameters'

    def insert(self, timestamp, region, minute_length, minute_offset, grid_baselength, count_threshold, count):
        sql = 'INSERT INTO ' + self.full_table_name + ' ' + \
              '("timestamp", region, minute_length, minute_offset, grid_baselength, count_threshold, "count")' + \
              'VALUES (%(timestamp)s, %(region)s, %(minute_length)s, %(minute_offset)s, %(grid_baselength)s, %(count_threshold)s, %(count)s);'

        parameters = {
            'timestamp': timestamp,
            'region': region,
            'minute_length': minute_length,
            'minute_offset': minute_offset,
            'grid_baselength': grid_baselength,
            'count_threshold': count_threshold,
            'count': count
        }

        self.execute(sql, parameters)



# -*- coding: utf8 -*-

"""

   Copyright 2014-2022 Andreas Würl

   Licensed under the Apache License, Version 2.0 (the "License");
   you may not use this file except in compliance with the License.
   You may obtain a copy of the License at

       http://www.apache.org/licenses/LICENSE-2.0

   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   See the License for the specific language governing permissions and
   limitations under the License.

"""

from .. import data, Error


class BuilderError(Error):
    pass


class Base:
    pass


class Timestamp(Base):
    def __init__(self):
        super().__init__()
        self.timestamp = None

    def set_timestamp(self, timestamp, nanosecond=0):
        if not timestamp:
            self.timestamp = None
        elif isinstance(timestamp, data.Timestamp):
            self.timestamp = timestamp + nanosecond
        else:
            self.timestamp = data.Timestamp(timestamp, nanosecond=nanosecond)
        return self

    def build(self):
        return self.timestamp


class Event(Timestamp):
    def __init__(self):
        super().__init__()
        self.x_coord = 0
        self.y_coord = 0

    def set_x(self, x_coord):
        self.x_coord = x_coord
        return self

    def set_y(self, y_coord):
        self.y_coord = y_coord
        return self

    def build(self):
        return data.Event(self.timestamp, self.x_coord, self.y_coord)



# -*- coding: utf8 -*-

"""

   Copyright 2014-2022 Andreas Würl

   Licensed under the Apache License, Version 2.0 (the "License");
   you may not use this file except in compliance with the License.
   You may obtain a copy of the License at

       http://www.apache.org/licenses/LICENSE-2.0

   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   See the License for the specific language governing permissions and
   limitations under the License.

"""

import re

from .base import Base, Event, BuilderError
from .. import data


class Station(Event):
    """
    class for building of station objects
    """

    station_parser = re.compile(r'station;(\d+)')
    user_parser = re.compile(r'user;(\d+)')
    city_parser = re.compile(r'city;"([^"]+)"')
    country_parser = re.compile(r'country;"([^"]+)"')
    position_parser = re.compile(r'pos;([-0-9.]+);([-0-9.]+);([-0-9.]+)')
    status_parser = re.compile(r'status;"?([^ ]+)"?')
    board_parser = re.compile(r'board;"?([^ ]+)"?')
    last_signal_parser = re.compile(r'last_signal;"([-: 0-9]+)" ?')

    def __init__(self):
        super().__init__()
        self.number = -1
        self.user = -1
        self.name = None
        self.country = None
        self.status = None
        self.board = None

    def set_number(self, number):
        self.number = number
        return self

    def set_user(self, user):
        self.user = user
        return self

    def set_name(self, name):
        self.name = name
        return self

    def set_country(self, country):
        self.country = country
        return self

    def set_board(self, board):
        self.board = board

    def set_status(self, status):
        self.status = status

    def from_line(self, line):
        try:
            self.set_number(int(self.station_parser.findall(line)[0]))
            self.set_user(int(self.user_parser.findall(line)[0]))
            self.set_name(self.city_parser.findall(line)[0])
            self.set_country(self.country_parser.findall(line)[0])
            pos = self.position_parser.findall(line)[0]
            self.set_x(float(pos[1]))
            self.set_y(float(pos[0]))
            self.set_board(self.board_parser.findall(line)[0])
            # self.set_status(self.status_parser.findall(line)[0])
            self.set_timestamp(self.last_signal_parser.findall(line)[0])
        except (KeyError, ValueError, IndexError) as e:
            raise BuilderError(e)
        return self

    def build(self):
        return data.Station(self.number, self.user, self.name, self.country,
                            self.x_coord, self.y_coord, self.timestamp, self.status,
                            self.board)


class StationOffline(Base):
    def __init__(self):
        super().__init__()
        self.id_value = -1
        self.number = -1
        self.begin = None
        self.end = None

    def set_id(self, id_value):
        self.id_value = id_value
        return self

    def set_number(self, number):
        self.number = number
        return self

    def set_begin(self, begin):
        self.begin = begin
        return self

    def set_end(self, end):
        self.end = end
        return self

    def build(self):
        return data.StationOffline(self.id_value, self.number, self.begin, self.end)



# -*- coding: utf8 -*-

"""

   Copyright 2014-2016 Andreas Würl

   Licensed under the Apache License, Version 2.0 (the "License");
   you may not use this file except in compliance with the License.
   You may obtain a copy of the License at

       http://www.apache.org/licenses/LICENSE-2.0

   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   See the License for the specific language governing permissions and
   limitations under the License.

"""

import logging
import os

from injector import singleton, inject

from .base import HttpFileTransport, BlitzortungDataPath, BlitzortungDataPathGenerator
from .. import builder


@singleton
class RawSignalsBlitzortungDataProvider:
    logger = logging.getLogger(__name__)

    @inject
    def __init__(self, data_transport: HttpFileTransport, data_url: BlitzortungDataPath,
                 url_path_generator: BlitzortungDataPathGenerator, waveform_builder: builder.RawWaveformEvent):
        self.data_transport = data_transport
        self.data_url = data_url
        self.url_path_generator = url_path_generator
        self.waveform_builder = waveform_builder

    def get_raw_data_since(self, latest_data, region, station_id):
        self.logger.debug("import raw data since %s" % latest_data)

        for url_path in self.url_path_generator.get_paths(latest_data):
            target_url = self.data_url.build_path(
                os.path.join(str(station_id), url_path),
                region=region,
                host_name='signals')

            for line in self.data_transport.read_lines(target_url):
                try:
                    yield self.waveform_builder.from_string(line).build()
                except builder.BuilderError as e:
                    self.logger.warn(str(e))



# -*- coding: utf8 -*-

"""

   Copyright 2014-2016 Andreas Würl

   Licensed under the Apache License, Version 2.0 (the "License");
   you may not use this file except in compliance with the License.
   You may obtain a copy of the License at

       http://www.apache.org/licenses/LICENSE-2.0

   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   See the License for the specific language governing permissions and
   limitations under the License.

"""

import itertools

from injector import inject

from .base import Event, BuilderError
from .. import data


class ChannelWaveform:
    """
    class for building of waveforms within raw signal objects
    """
    fields_per_channel = 11

    def __init__(self):
        self.channel_number = None
        self.amplifier_version = None
        self.antenna = None
        self.gain = None
        self.values = 0
        self.start = None
        self.bits = 0
        self.shift = None
        self.conversion_gap = None
        self.conversion_time = None
        self.waveform = None

    def from_field_iterator(self, field):
        self.channel_number = int(next(field))
        self.amplifier_version = next(field)
        self.antenna = int(next(field))
        self.gain = next(field)
        self.values = int(next(field))
        self.start = int(next(field))
        self.bits = int(next(field))
        self.shift = int(next(field))
        self.conversion_gap = int(next(field))
        self.conversion_time = int(next(field))
        self.__extract_waveform_from_hex_string(next(field))
        return self

    def __extract_waveform_from_hex_string(self, waveform_hex_string):
        hex_character = iter(waveform_hex_string)
        self.waveform = self.values * [0]
        bits_per_char = 4
        if self.bits == 0:
            self.bits = len(waveform_hex_string) // self.values * bits_per_char
        chars_per_sample = self.bits // bits_per_char
        value_offset = -(1 << (chars_per_sample * 4 - 1))

        for index in range(0, self.values):
            value_text = "".join(itertools.islice(hex_character, chars_per_sample))
            value = int(value_text, 16)
            self.waveform[index] = value + value_offset

    def build(self):
        return data.ChannelWaveform(
            self.channel_number,
            self.amplifier_version,
            self.antenna,
            self.gain,
            self.values,
            self.start,
            self.bits,
            self.shift,
            self.conversion_gap,
            self.conversion_time,
            self.waveform)


class RawWaveformEvent(Event):
    """
    class for building of raw signal objects
    """

    @inject
    def __init__(self, channel_builder: ChannelWaveform):
        super().__init__()
        self.altitude = 0
        self.channels = []

        self.channel_builder = channel_builder

    def build(self):
        return data.RawWaveformEvent(
            self.timestamp,
            self.x_coord,
            self.y_coord,
            self.altitude,
            self.channels
        )

    def set_altitude(self, altitude):
        self.altitude = altitude
        return self

    def from_json(self, json_object):
        self.set_timestamp(json_object[0])
        self.set_x(json_object[1])
        self.set_y(json_object[2])
        self.set_altitude(json_object[3])
        if len(json_object[9]) > 1:
            self.set_y(json_object[9][1])

        return self

    def from_string(self, string):
        """ Construct strike from blitzortung.org text format data line """
        if string:
            try:
                field = iter(string.split(' '))
                self.set_timestamp(next(field) + ' ' + next(field))
                self.timestamp.datetime += datetime.timedelta(seconds=1)
                self.set_y(float(next(field)))
                self.set_x(float(next(field)))
                self.set_altitude(int(next(field)))

                self.channels = []
                while True:
                    try:
                        self.channel_builder.from_field_iterator(field)
                    except StopIteration:
                        break
                    self.channels.append(self.channel_builder.build())
            except ValueError as e:
                raise BuilderError(e, string)

        return self



# -*- coding: utf8 -*-

"""

   Copyright 2014-2016 Andreas Würl

   Licensed under the Apache License, Version 2.0 (the "License");
   you may not use this file except in compliance with the License.
   You may obtain a copy of the License at

       http://www.apache.org/licenses/LICENSE-2.0

   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   See the License for the specific language governing permissions and
   limitations under the License.

"""

import gzip
import io
import logging

from injector import singleton, inject

from .base import HttpFileTransport, BlitzortungDataPath
from .. import builder


@singleton
class StationsBlitzortungDataProvider:
    logger = logging.getLogger(__name__)

    @inject
    def __init__(self, data_transport: HttpFileTransport, data_url: BlitzortungDataPath,
                 station_builder: builder.Station):
        self.data_transport = data_transport
        self.data_url = data_url
        self.station_builder = station_builder

    def get_stations(self, region=1):
        current_stations = []
        target_url = self.data_url.build_path('Protected/stations.txt.gz', region=region)
        for station_line in self.data_transport.read_lines(target_url, post_process=self.pre_process):
            try:
                current_stations.append(self.station_builder.from_line(station_line).build())
            except builder.BuilderError:
                self.logger.debug("error parsing station data '%s'" % station_line)
        return current_stations

    @staticmethod
    def pre_process(data):
        data = io.BytesIO(data)
        out_data = gzip.GzipFile(fileobj=data).read()
        return out_data



# -*- coding: utf8 -*-

"""

   Copyright 2014-2016 Andreas Würl

   Licensed under the Apache License, Version 2.0 (the "License");
   you may not use this file except in compliance with the License.
   You may obtain a copy of the License at

       http://www.apache.org/licenses/LICENSE-2.0

   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   See the License for the specific language governing permissions and
   limitations under the License.

"""

import datetime
import glob
import json
import os
import subprocess

from blitzortung.data import Timestamp
from . import builder


class Raw:
    BO_DATA_EXECUTABLE = 'bo-data'

    def __init__(self, file_path):
        self.file_path = file_path

    def get_file_path(self):
        return self.file_path

    def get_file_name(self):
        return os.path.basename(self.file_path)

    def get_data(self, start_time=None, end_time=None):
        return [builder.RawEvent().from_json(element).build()
                for element in self.__execute(start_time, end_time)]

    def get_waveform_data(self, start_time=None, end_time=None):
        return [builder.RawWaveformEvent().from_json(element).build()
                for element in self.__execute(start_time, end_time, '--long-data')]

    def get_info(self, start_time=None, end_time=None):
        return self.__execute(start_time, end_time, '--mode', 'info')

    def get_histogram(self, start_time=None, end_time=None):
        return self.__execute(start_time, end_time, '--mode', 'histogram')

    def __repr__(self):
        return "files.Raw(%s)" % (os.path.basename(self.file_path))

    def __execute(self, start_time, end_time, *additional_args):
        args = [self.BO_DATA_EXECUTABLE, '-j', '-i', self.file_path]
        if start_time:
            args += ['-s', start_time]
        if end_time:
            args += ['-e', end_time]
        data_pipe = subprocess.Popen(args + list(additional_args), stdout=subprocess.PIPE)
        (output, _) = data_pipe.communicate()
        return json.loads(output)


class RawFile:
    def __init__(self, config):
        raw_file_names = glob.glob(os.path.join(config.get_raw_path(), '*.bor'))

        raw_file_names.sort()

        self.raw_files = {}

        for raw_file_name in raw_file_names:
            try:
                date = datetime.datetime.strptime(raw_file_name[-12:-4], '%Y%m%d').date()
            except ValueError:
                continue
            if date not in self.raw_files:
                self.raw_files[date] = raw_file_name
            else:
                raise ValueError("ERROR: double date! " + raw_file_name + " vs. " + self.raw_files[date])

    def get(self, date):
        if date in self.raw_files:
            return self.raw_files[date]
        else:
            raise KeyError("no file for date " + date.strftime('%Y-%m-%d'))

    def get_dates(self):
        dates = self.raw_files.keys()
        dates.sort()
        return dates


class Archive:
    def __init__(self, config):
        self.dates_filecount = {}

        self.root_path = config.get_archive_path()
        root_depth = self.__get_path_depth(self.root_path)

        for current_path, dirs, files in os.walk(self.root_path):
            depth = self.__get_path_depth(current_path) - root_depth

            if depth == 3:
                date_string = "-".join(self.__split_path_into_components(current_path)[-depth:])
                self.dates_filecount[Timestamp(date_string)] = len(files)

    def get_dates_filecount(self):
        return self.dates_filecount

    def get_files_for_date(self, date_string):
        result = []
        date = Timestamp(date_string)
        if date in self.dates_filecount:

            for file_path in glob.glob(os.path.join(self.__get_path_for_date(date), '*')):
                result.append(Raw(file_path))

        return result

    def __get_path_for_date(self, date):
        path = self.root_path

        for format_string in ['%Y', '%m', '%d']:
            path = os.path.join(path, date.strftime(format_string))

        return path

    def __get_path_depth(self, path):
        return len(self.__split_path_into_components(path))

    def __split_path_into_components(self, path):
        (rest, last) = os.path.split(path)
        if last == "":
            return []
        else:
            components = self.__split_path_into_components(rest)
            components.append(last)
            return components


class Data:
    def __init__(self, raw_file_path, time):
        self.raw_file_path = raw_file_path
        self.time = time
        self.error = False

    def get(self, long_format=False):
        start = self.time.get_start_time()
        start_time = start.strftime("%H%M")

        end = self.time.get_end_minute()
        end_time = end.strftime("%H%M")

        self.error = False

        raw_file = self.raw_file_path.get_paths(start.date())

        if long_format:
            return self.get_output(raw_file, start_time, end_time, True)
        else:
            return self.get_data(raw_file, start_time, end_time)

    @staticmethod
    def get_output(raw_file, start_time, end_time, long_format=False):
        cmd = ['bo-data', '-i', raw_file, '-s', start_time, '-e', end_time]
        if long_format:
            cmd.append('--long-data')
        data_pipe = subprocess.Popen(cmd, stdout=subprocess.PIPE)
        (output, _) = data_pipe.communicate()

        return output.splitlines(keepends=False)

    def get_data(self, raw_file, start_time, end_time):
        raw_events = []

        for line in self.get_output(raw_file, start_time, end_time):
            raw_event_builder = builder.RawEvent()
            raw_event_builder.from_string(line)
            raw_events.append(raw_event_builder.build())

        return raw_events

    def list(self):
        for event in self.get():
            print(event)

    def list_long(self):
        for line in self.get(True):
            print(line)
